#include <stdint.h> /* Standard Integer Types */
#include <stdio.h> /* Standard IO facilities */
#include <stdlib.h> /* General utilities */
#include <string.h> /* Strings */
#include <avr/io.h> /* AVR device-specific IO definitions */
#include <util/delay.h> /* Convenience functions for busy-wait delay loops */
#include <avr/interrupt.h> /* Interrupts */

#include "main.h"
#include "uart.h"

/* global variables */
volatile uint8_t pwr_ok;
volatile uint8_t fan;
volatile uint16_t tachometer = 0;
volatile uint8_t timer2_ovf = 0;

const uint16_t TIMER2_PRESCALE[8] = {0,1,8,32,64,128,256,1024};

/* PWR_OK interrupt */
ISR(PCINT0_vect) { /* PCI0 Interrupt Vector for PCINT[7:0] */
	if (pwr_ok!=(PINB&(1<<PWR_OK))) { /* did the PWR_OK pin state changed */
		pwr_ok = PINB&(1<<PWR_OK); /* save new state */
		if (pwr_ok) {
			puts("power ok");
		} else {
			puts("power ko");
		}
	}
}

/* fan tachometer interrupt */
ISR(PCINT1_vect) { /* PCI0 Interrupt Vector for PCINT[14:8] */
	if (fan!=(PINC&(1<<FAN))) { /* did the FAN pin state changed */
		fan = PINC&(1<<FAN); /* save new state */
		if (fan) {
			tachometer = timer2_ovf*256+TCNT2; /* save time */
			TCNT2 = 0; /* reset timer 2 */
			timer2_ovf = 0; /* reset timer 2 overflow counter */
		}
	}
}

/* timer 2 interrupt used to measure fan tachometer */
ISR(TIMER2_OVF_vect) { /* timer 2 overflow interrupt verctor */
	if(pwr_ok) { /* only measure tachometer if power is ok */
		timer2_ovf++; /* increase time 2 overflow counter */
	}
}

static void ioinit(void)
{
	/* configure power */
	DDRB |= (1<<nPS_ON); /* nPS_ON is output */
	PORTB |= (1<<nPS_ON); /* switch off power supply */
	DDRB &= ~(1<<PWR_OK); /* PWR_ON (PB1/PCINT1) is input */
	pwr_ok = PINB&(1<<PWR_OK); /* save state */
	PCIFR &= ~(1<<PCIF0); /* clear interrupt flag */
	PCICR |= (1<<PCIE0); /* enable interrupt for PCINT[7:0] */
	PCMSK0 |= (1<<PCINT1); /* enable interrupt for PCINT1 */

	/* configure peripherals */
	DDRB &= ~(1<<IR); /* IR receiver is input */

	/* configure FAN */
	DDRC &= ~(1<<FAN); /* FAN (PC5/PCINT13) is input */
	fan = PINC&(1<<FAN); /* save state */
	PCIFR &= ~(1<<PCIF1); /* clear interrupt flag */
	PCICR |= (1<<PCIE1); /* enable interrupt for PCINT[14:8] */
	PCMSK1 |= (1<<PCINT13); /* enable interrupt for PCINT1 */
	/* use timer 2 to measure the tachometer */
	/* use normal mode */
	TCCR2A &= ~((1<<WGM20)|(1<<WGM21));
	TCCR2B &= ~(1<<WGM22);
	/* /256 prescale timer */
	TCCR2B |= (1<<CS22)|(1<<CS21);
	TCCR2B &= ~(1<<CS20);
	TIFR2 &= ~(1<<TOV2); /* clear timer 2 overflow interrupt flag */
	TIMSK2 |= (1<<TOIE2); /* enable timer 2 overflow interrupt */

	/* configure channels (used for powering LEDs using an nMOS) */
	DDRC |= (1<<CH1_1)|(1<<CH1_2)|(1<<CH1_3)|(1<<CH1_4)|(1<<CH1_5); /* CH1_x is output */
	PORTC &= ~((1<<CH1_1)|(1<<CH1_2)|(1<<CH1_3)|(1<<CH1_4)|(1<<CH1_5)); /* switch off CH1_x */
	DDRD |= (1<<CH2_1)|(1<<CH2_2)|(1<<CH2_3)|(1<<CH2_4)|(1<<CH2_5); /* CH2_x is output */
	PORTD &= ~((1<<CH2_1)|(1<<CH2_2)|(1<<CH2_3)|(1<<CH2_4)|(1<<CH2_5)); /* switch off CH2_x */

	/* use UART as terminal */
	uart_init();
	stdout = &uart_output;
	stdin  = &uart_input;

	/* configure LED (on PD6/OC0A) */
	DDRD |= (1<<LED); /* LED is output */
	/* use phase correct PWM mode (because fast PWM is always on for at least 1 cycle) */
	TCCR0A &= ~(1<<WGM01);
	TCCR0A |= (1<<WGM00);
	TCCR0B &= ~(1<<WGM02);
	/* /64 prescale timer */
	TCCR0B &= ~(1<<CS02);
	TCCR0B |= (1<<CS01)|(1<<CS00);
	/* clear on match, because the pin is used as sink */
	TCCR0A |= (1<<COM0A1)|(1<<COM0A0);
	/* start with LED on */
	OCR0A = 0xff;

	sei(); /* enable interrupts */
}

int main(void)
{
	ioinit(); /* initialize IOs */

	char input;

	puts("Hello world!");
	_delay_ms(1000);
	while(1) {
		input = getchar();
		printf("%c\n", input);
		switch (input) {
			case 'l':
				PIND |= (1<<LED);
				printf("LED: ");
				if (PIND&(1<<LED)) {
					puts("off");
				} else {
					puts("on");
				}
				break;
			case 'a':
				printf("switching power supply ");
				if (PINB&(1<<nPS_ON)) {
					puts("on");
				} else {
					puts("off");
				}
				PINB |= (1<<nPS_ON);
				break;
			case 's':
				printf("power: ");
				if (PINB&(1<<PWR_OK)) {
					puts("ok");
				} else {
					puts("ko");
				}
				break;
			case '1':
				PINC |= (1<<CH1_1);
				printf("CH1_1: ");
				if (PINC&(1<<CH1_1)) {
					puts("on");
				} else {
					puts("off");
				}
				break;
			case '2':
				PINC |= (1<<CH1_2);
				printf("CH1_2: ");
				if (PINC&(1<<CH1_2)) {
					puts("on");
				} else {
					puts("off");
				}
				break;
			case '3':
				PINC |= (1<<CH1_3);
				printf("CH1_3: ");
				if (PINC&(1<<CH1_3)) {
					puts("on");
				} else {
					puts("off");
				}
				break;
			case '4':
				PINC |= (1<<CH1_4);
				printf("CH1_4: ");
				if (PINC&(1<<CH1_4)) {
					puts("on");
				} else {
					puts("off");
				}
				break;
			case '5':
				PINC |= (1<<CH1_5);
				printf("CH1_5: ");
				if (PINC&(1<<CH1_5)) {
					puts("on");
				} else {
					puts("off");
				}
				break;
			case '6':
				PIND |= (1<<CH2_1);
				printf("CH2_1: ");
				if (PIND&(1<<CH2_1)) {
					puts("on");
				} else {
					puts("off");
				}
				break;
			case '7':
				PIND |= (1<<CH2_2);
				printf("CH2_2: ");
				if (PIND&(1<<CH2_2)) {
					puts("on");
				} else {
					puts("off");
				}
				break;
			case '8':
				PIND |= (1<<CH2_3);
				printf("CH2_3: ");
				if (PIND&(1<<CH2_3)) {
					puts("on");
				} else {
					puts("off");
				}
				break;
			case '9':
				PIND |= (1<<CH2_4);
				printf("CH2_4: ");
				if (PIND&(1<<CH2_4)) {
					puts("on");
				} else {
					puts("off");
				}
				break;
			case '0':
				PIND |= (1<<CH2_5);
				printf("CH2_5: ");
				if (PIND&(1<<CH2_5)) {
					puts("on");
				} else {
					puts("off");
				}
				break;
			case '-':
				if (OCR0A>0) {
					OCR0A--;
				}
				printf("decreasing LED: %d\n",OCR0A);
				break;
			case '+':
				if (OCR0A<0xff) {
					OCR0A++;
				}
				printf("increasing LED: %d\n",OCR0A);
				break;
			case 't':
				if (tachometer) {
					uint16_t prescale = TIMER2_PRESCALE[TCCR2B&((1<<CS22)|(1<<CS21)|(1<<CS20))];
					if (prescale) {
						uint32_t speed = ((60*F_CPU)/(prescale*(uint32_t)tachometer))/2;
						printf("fan speed: %lurpm\n",speed);
						printf("fan speed: %lurpm\n",timer2_ovf);
					}
				} else {
					printf("fan off or too slow\n");
				}
				break;
		}
	}

	return 0;
}
